"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const validate_1 = require("@code-engine/validate");
const events_1 = require("events");
const ono_1 = require("ono");
const os = require("os");
const worker_1 = require("./worker");
/**
 * Runs CodeEngine plugins on worker threads.
 */
class WorkerPool extends events_1.EventEmitter {
    constructor(concurrency, context) {
        super();
        /** @internal */
        this._workers = [];
        /** @internal */
        this._isDisposed = false;
        /** @internal */
        this._moduleCounter = 0;
        /** @internal */
        this._roundRobinCounter = 0;
        concurrency = validate_1.validate.number.integer.positive(concurrency, "concurrency", os.cpus().length);
        validate_1.validate.type.object(context, "context");
        this._cwd = validate_1.validate.string.nonWhitespace(context.cwd, "context.cwd");
        this._createWorkers(concurrency, context);
    }
    /**
     * Indicates the number of worker threads in the pool.
     */
    get size() {
        return this._workers.length;
    }
    /**
     * Indicates whether the `dispose()` method has been called.
     * Once disposed, the `WorkerPool` instance is no longer usable.
     */
    get isDisposed() {
        return this._isDisposed;
    }
    /**
     * Imports the specified `FileProcessor` module in all worker threads.
     *
     * @returns - A proxy function that executes the processor in one of the threads.
     */
    async importFileProcessor(module) {
        this._assertNotDisposed();
        let { moduleId, data } = normalizeModule(module);
        let cwd = this._cwd;
        let moduleUID = ++this._moduleCounter;
        let message = { type: "importFileProcessor", cwd, moduleUID, moduleId, data };
        // Import the JavaScript module in all worker threads
        let [name] = await Promise.all(this._workers.map((worker) => worker.importFileProcessor(message)));
        // Create a CodeEngine FileProcessor function that executes the module on a worker thread
        let plugin = {
            [name]: (file, context) => {
                // Select a worker from the pool to process the files
                let worker = this._select();
                // Process the file on the worker thread
                return worker.processFile(moduleUID, file, context);
            }
        };
        // Return the FileProcessor function with the same name as the one in the module
        return plugin[name];
    }
    /**
     * Imports the specified JavaScript module in all worker threads.
     */
    async importModule(module) {
        this._assertNotDisposed();
        let { moduleId, data } = normalizeModule(module);
        let cwd = this._cwd;
        let message = { type: "importModule", cwd, moduleId, data };
        // Import the JavaScript module in all worker threads
        await Promise.all(this._workers.map((worker) => worker.importModule(message)));
    }
    /**
     * Terminates all worker threads.
     */
    async dispose() {
        this._isDisposed = true;
        let workers = this._workers;
        this._workers = [];
        await Promise.all(workers.map((worker) => worker.terminate()));
    }
    /**
     * Creates the specified number of worker threads.
     * @internal
     */
    _createWorkers(concurrency, context) {
        // Re-emit all errros from workers
        let emitError = (error) => this.emit("error" /* Error */, error);
        for (let i = 0; i < concurrency; i++) {
            let worker = new worker_1.Worker(context);
            worker.on("error", emitError);
            this._workers.push(worker);
        }
    }
    /**
     * Selects a `Worker` from the pool to perform a task.
     * @internal
     */
    _select() {
        // For now, we just use a simple round-robin strategy,
        // but we may employ a more advanced selection strategy later
        return this._workers[this._roundRobinCounter++ % this._workers.length];
    }
    /**
     * Throws an error if the `WorkerPool` has been disposed.
     * @internal
     */
    _assertNotDisposed() {
        if (this.isDisposed) {
            throw ono_1.ono(`CodeEngine cannot be used after it has been disposed.`);
        }
    }
}
exports.WorkerPool = WorkerPool;
/**
 * Normalizes a module definition.
 */
function normalizeModule(module) {
    if (typeof module === "string") {
        return { moduleId: module };
    }
    else {
        return module;
    }
}
//# sourceMappingURL=worker-pool.js.map